#include <stdio.h>
#include "config.h"
#include "area.h"
#include "status.h"
#include "message.h"
#include "command.h"

#define SYMBOL_FLAG_NUM 2
typedef struct {
    noun_id_t nid;
    flag_id_t fids[SYMBOL_FLAG_NUM];
    char *description;
    char *message;
    char *hint;
} symbol_t;


typedef struct {
    int num;
    symbol_t *symbols;
} symbol_entry_t;
    
symbol_t death_symbols[] = {
    {ni_glasses,   
     {fi_death_by_touching_water, fi_invalid},
     "襷ζδȤˤĤƤϤޤͤʤ",
     "ο˿ƻह",
     "򾷤ҥ"},
    {ni_necklace,  
     {fi_death_by_touching_shadow, fi_invalid},
     "襷׿ͤμ",
     "2αƤ˿ƻह",
     "̵˿Ȼ򾷤"},
    {ni_armor,     
     {fi_death_by_armor, fi_invalid},
     "Υߥ˥奢",
     "ڤƻह",
     "ϼϼƤʤ"},
    {ni_slingshot,
     {fi_death_by_exiting_window, fi_invalid},
     "ʴǤݤäޤ줿Ѥ󤳤",
     "˶ޤƻह",
     "Фʤ衢нФʤ"},
    {ni_skull,
     {fi_death_by_starvation, fi_invalid},
     "ï",
     "ޥɥʥɤŲƧߤƻह",
     "ɤʤ˿ȡ"},
    {ni_tentacle,
     {fi_death_by_tentacle, fi_invalid},
     "±β­ǤϤʤ",
     "ۼԤƻह",
     "͸ˤ줿"},
};
#define DEATH_SYMBOL_NUM (sizeof(death_symbols)/sizeof(symbol_t))

symbol_t alternative_symbols[] = {
    {ni_fossil,
     {fi_clear_armor_alive, fi_clear_armor_dead},
     "ɥꥻ饹ʤΥʥȤβФ"
     "餯ܤ̳ƻ"
     "Υ塼ǻФ¸֤ˤ褤"
     "䤬Ťƥ꡼˥󥰤ѤǤ"
     "ʤˤɤǤ⤤ƬήǤ롣",
     "ݤơݤ˥ꥢ",
     "ݤʤƤ褤"},
    {ni_billiard_ball,
     {fi_clear_treasure_chest_unlocked, fi_clear_treasure_chest_locked},
     "ӥ䡼ɥܡ",
     "Ȣξ򳫤ơƥꥢ",
     "Ȣθϼ˺Ƥʤ"},
    {ni_portrait,
     {fi_clear_portrait_enabled, fi_clear_portrait_disabled},
     "ܤưޥɥʥšԻ׵ĤΰĤ餷",
     "򳰤ơ˥ꥢ",
     "褯ѻȡ"},
    {ni_human_bone,
     {fi_clear_using_human_bone, fi_clear_using_chicken_bone},
     "ïܹ",
     "ͤιܤιǥȥ󵡴ؤưƥꥢ",
     "륷ʤ鲿Ǥ⤤"},
    {ni_telephone,
     {fi_clear_phone_call_answered, fi_clear_phone_call_ignored},
     "뼰ä",
     "ä˽ФơФ˥ꥢ",
     "ä򤫤ƤïФ"},
};
#define ALTERNATIVE_SYMBOL_NUM (sizeof(alternative_symbols)/sizeof(symbol_t))

symbol_t complete_symbols[] = {
    {ni_lantern,   
     {fi_clear_map_completed, fi_invalid},
     "äڤ򵯤ʤԻ׵Ĥʥ󥿥",
     "ޥå򤹤٤Ʋä",
     "ޤʤ⤱"},
    {ni_girl_doll,   
     {fi_clear_dummy_item_completed, fi_invalid},
     "̣ꤲǰ̣Τʤͷ",
     "ߡƥ򤹤٤",
     "ΤϤʤǤ"},
    {ni_witch_doll,   
     {fi_clear_read_completed, fi_invalid},
     "ˤߡäݤοͷ",
     "ʸ򤹤٤ɤ",
     "ɤΤϤʤǤɤ"},
    {ni_large_robot,   
     {fi_clear_character_completed, fi_invalid},
     "Ƥβ˸ʹΥܥåȤδ",
     "طԤλѤ򤹤٤Ƹ",
     "ɤΤϤʤǤɤ"},
};
#define COMPLETE_SYMBOL_NUM (sizeof(complete_symbols)/sizeof(symbol_t))

symbol_t progress_symbols[] = {
    {ni_flowerpot,   
     {fi_death_entered_house, fi_invalid},
     "ȥեɤȭˤϤƤ롣",
     "ߤä",
     "ʤȤƸؤ򳫤褦"},
    {ni_fishing_rod,   
     {fi_death_entered_2f, fi_invalid},
     "μ̣Ȥ",
     "2ä",
     "1Τɤ˸"},
    {ni_candle,   
     {fi_death_candle_lit, fi_invalid},
     "ʤʤФĤʤϹ",
     "Ѥ",
     "ƲˤƵԤ٤"},
    {ni_bottle,   
     {fi_death_magic_circle_invoked, fi_invalid},
     "ᤤӤˤϥФʴäƤ롣",
     "ˡؤư",
     "򸫤ĤԤ٤"},
    {ni_amulet,   
     {fi_clear_game, fi_invalid},
     "ΰ",
     "ǥ󥰤ޤ",
     "ۤʤȤæФ褦"},
};
#define PROGRESS_SYMBOL_NUM (sizeof(progress_symbols)/sizeof(symbol_t))

symbol_entry_t symbol_table[] = {
    {DEATH_SYMBOL_NUM, death_symbols},
    {ALTERNATIVE_SYMBOL_NUM, alternative_symbols},
    {COMPLETE_SYMBOL_NUM, complete_symbols},
    {PROGRESS_SYMBOL_NUM, progress_symbols},
};

#define SYMBOLS_NUM (sizeof(symbol_table)/sizeof(symbol_entry_t))

int is_achievements_on(flag_id_t fids[]);
int count_achievements(int num, symbol_t symbols[]);
void message_active_symbols(int num, symbol_t symbols[]);
void message_achievements_count(char *header, int num, symbol_t symbols[]);
char *achievements_hint(int num, symbol_t symbols[]);
flag_id_t *symbol_nid2fids(noun_id_t nid, int num, symbol_t symbols[]);
char *symbol_nid2message(noun_id_t nid, int num, symbol_t symbols[]);
void message_symbol(noun_id_t nid, int num, symbol_t symbols[]);
void tentacle_duck(void);
void tentacle_attack(void);
int command_symbols(verb_id_t vid, noun_id_t nid, int *exist);
    
int area_achievements(verb_id_t vid, noun_id_t nid, int *exist){
    assert(vid != vi_invalid);
	*exist = 1;
    switch(nid){
    case ni_invalid:
        switch(vid){
        case vi_look:
            DescribeArea;
			message("ɺݤˤϤ줾%nl%nl롣"
                    "̤ˤ%nlꡢμˤ%nl롣",
					ni_cabinet, ni_table, ni_fireplace, ni_treasure_chest);
            if (Not(achievements_dog_sitting)){
                message("%nwα¦ˤ%nlꡢ%nlΩäƤ롣",
                        ni_fireplace, ni_door, ni_dog);
            }
            else {
                message("%nnα¦ˤ%nl롣",
                        ni_fireplace, ni_door);
                message("ξˤ%nl¤äƤ롣",
                        ni_dog);
            }
            message("ȥ̤ˤϡ%vפϤƤ",
                    vi_backward);
            return 1;
        case vi_backward:
        case vi_down:
		case vi_exit:
            move_to_title();
            return 1;
        default:
            break;
        }
        break;
    case ni_table:
        switch (vid){
        case vi_look:
            message("ϻΥơ֥");
            message("ơ֥ξˤ\\");
            message_active_symbols(DEATH_SYMBOL_NUM, death_symbols);
            return 1;
        case vi_examine:
            message_achievements_count("", DEATH_SYMBOL_NUM, death_symbols);
            return 1;
        default:
            break;
        }
        break;
    case ni_fireplace:
        switch (vid){
        case vi_look:
            message("ϧ");
            message("ˤ\\");
            message_active_symbols(ALTERNATIVE_SYMBOL_NUM, alternative_symbols);
            return 1;
        case vi_examine:
            message_achievements_count("", ALTERNATIVE_SYMBOL_NUM, 
                                       alternative_symbols);
            return 1;
        default:
            break;
        }
        break;
    case ni_treasure_chest:
        switch (vid){
        case vi_look:
            message("ϤꤳߤȢ");
            if (Is(achievements_treasure_chest_opened) &&
                Not(achievements_duck_moved))
            {
                message("Ȣˤ%nläƤ롣", ni_rubber_duck);
            }
            message("դˤ\\");
            message_active_symbols(COMPLETE_SYMBOL_NUM, complete_symbols);
            return 1;
        case vi_examine:
            message_achievements_count("ꤳ", COMPLETE_SYMBOL_NUM,
                                       complete_symbols);
            return 1;
        case vi_open:
            if (Not(achievements_treasure_chest_opened)){
                if (Not(achievements_treasure_chest_unlocked)){
                    message("ݤäƤ롣");
                    return 1;
                }
                message("%nn򳫤", nid);
                Set(achievements_treasure_chest_opened);
                if (Not(achievements_chicken_dropped)){
                    message("Ȣ椫%nlФƤ", ni_chicken);
                    Set(achievements_chicken_dropped);
                    message("%nnäݤ򿶤ä%nnӤĤ",
                            ni_dog, ni_chicken);
                    Set(achievements_dog_sitting);
                }
            }
            else {
                message("%nwϤ⤦Ƥ롣", nid);
            }
            return 1;
        case vi_close:
            if (Is(achievements_treasure_chest_opened)){
                message("%nwĤ᤿", nid);
                Reset(achievements_treasure_chest_opened);
            }
            else {
                message("%nwĤޤäƤ롣", nid);
            }
            return 1;
        case vi_hit:
        case vi_kick:
        case vi_destroy:
            message("ϸʤȳʤ褦");
            return 1;
        default:
            break;
        }
        break;
    case ni_cabinet:
        switch (vid){
        case vi_look:
            message("ϿĽΥӥͥåȤ");
            message("ӥͥåȤξˤ\\");
            message_active_symbols(PROGRESS_SYMBOL_NUM, progress_symbols);
            if (Is(achievements_cabinet_opened)){
                if (Not(achievements_wooden_key_taken) &&
                    Not(achievements_sponge_taken)){
                    message("ˤ%nl%nläƤ롣",
                            ni_sponge, ni_wooden_key);
                }
                else if (Not(achievements_sponge_taken)){
                    message("ˤ%nläƤ롣", ni_sponge);
                }
                else if (Not(achievements_wooden_key_taken)){
                    message("ˤ%nläƤ롣", ni_wooden_key);
                }
                else {
                    message("϶");
                }
            }
            return 1;
        case vi_examine:
            message_achievements_count("Ľ", PROGRESS_SYMBOL_NUM,
                                       progress_symbols);
            return 1;
		case vi_open:
            if (Not(achievements_cabinet_opened)){
                message("%nn򳫤", ni_cabinet);
                Set(achievements_cabinet_opened);
            }
            else {
                message("%nnϤ⤦Ƥ롣", ni_cabinet);
            }
            return 1;
		case vi_close:
            if (Is(achievements_cabinet_opened)){
                message("%nnĤ᤿", ni_cabinet);
                Reset(achievements_cabinet_opened);
            }
            else {
                message("%nnĤƤ롣", ni_cabinet);
            }
            return 1;
        default:
            break;
        }
        break;
        /*
    case ni_glasses:
    case ni_necklace:
    case ni_armor:
    case ni_slingshot:
    case ni_skull:
    case ni_tentacle:
        if (!is_achievements_on(symbol_nid2fids(nid, DEATH_SYMBOL_NUM,
                                                death_symbols)))
        {
            ReturnNotExist;
        }
        switch (vid){
        case vi_look:
        case vi_examine:
            message_symbol(nid, DEATH_SYMBOL_NUM, death_symbols);
            return 1;
        default:
            break;
        }
        break;
    case ni_fossil:
    case ni_billiard_ball:
    case ni_human_bone:
    case ni_telephone:
    case ni_portrait:
        if (!is_achievements_on(symbol_nid2fids(nid, ALTERNATIVE_SYMBOL_NUM,
                                                alternative_symbols)))
        {
            ReturnNotExist;
        }
        switch (vid){
        case vi_look:
        case vi_examine:
            message_symbol(nid, ALTERNATIVE_SYMBOL_NUM, alternative_symbols);
            return 1;
        default:
            break;
        }
        break;
    case ni_lantern:
    case ni_girl_doll:
    case ni_witch_doll:
        if (!is_achievements_on(symbol_nid2fids(nid, COMPLETE_SYMBOL_NUM,
                                                complete_symbols)))
        {
            ReturnNotExist;
        }
        switch (vid){
        case vi_look:
        case vi_examine:
            message_symbol(nid, COMPLETE_SYMBOL_NUM, complete_symbols);
            return 1;
        default:
            break;
        }
        break;
    case ni_flowerpot:
    case ni_fishing_rod:
    case ni_candle:
    case ni_bottle:
    case ni_amulet:
        if (!is_achievements_on(symbol_nid2fids(nid, PROGRESS_SYMBOL_NUM,
                                                progress_symbols)))
        {
            ReturnNotExist;
        }
        switch (vid){
        case vi_look:
        case vi_examine:
            message_symbol(nid, PROGRESS_SYMBOL_NUM, progress_symbols);
            return 1;
        default:
            break;
        }
        break;
        */
    case ni_door:
        switch (vid){
        case vi_look:
        case vi_examine:
            if (Is(achievements_door_opened)){
                message("ɥ%nlο꤬Ƥ롣",
                        ni_it);
                return 1;
            }
            break;
        case vi_open:
            if (Is(achievements_door_opened)){
                message("Ϥ⤦Ƥ롣");
            }
            else if (Not(achievements_dog_sitting)){
                message("%nn", ni_dog);
            }
            else {
                message("ʤ򳫤ȡ\\");
                Set(achievements_door_opened);
                if (Is(achievements_all_completed) &&
                    Not(achievements_magic_circle_erased))
                {
                    msg_tentacle_blocked();
                }
                else {
                    tentacle_attack();
                }
            }
            return 1;
        case vi_close:
            if (Is(achievements_door_opened)){
                message("ʤϤʤȤ򲡤ФƥɥĤ᤿");
                Reset(achievements_door_opened);
            }
            else {
                message("⤦ĤޤäƤ롣");
            }
            return 1;
        default:
            break;
        }
        break;
    case ni_it:
        if (Not(achievements_door_opened)) ReturnNotExist;
        switch (vid){
        case vi_look:
            message("Ҥο");
            message("ʤߤˡ֤֤󡢤ϤΡҤȤפάǤ롣");
            return 1;
            //        case vi_talk:
        case vi_hit:
        case vi_kick:
        case vi_fight:
        case vi_destroy:
            message("԰٤ϤƤ");
            return 1;
        default:
            break;
        }
        break;
    case ni_dog:
        switch (vid){
        case vi_look:
            if (Is(achievements_treasure_chest_opened)){
                message("%nnˤ餷Ƥ롣", ni_chicken);
            }
            else {
                message("뤫Τ褦ΩդäƤ롣");
            }
            return 1;
        case vi_talk:
            message("ʤϸä");
            message("Ϥ狼ä褦ʴ򤷤ʹƤ롣");
            return 1;
        case vi_touch:
            message("ʤϸʤǤ");
            message("ϵ褵ˤƤ롣");
            return 1;
		case vi_push:
		case vi_pull:
        case vi_remove:
        case vi_move:
            message("Ȥưʤ");
            return 1;
		case vi_hit:
		case vi_kick:
		case vi_fight:
            message("臘ʤ褦");
            return 1;
        default:
            break;
        }
        break;
    case ni_wooden_key:
        if (has_item(nid)){
            switch (vid){
            case vi_use:
				if (Not(achievements_treasure_chest_unlocked)){
					message("Ȣθ");
					Set(achievements_treasure_chest_unlocked);
					lose_item(nid);
				}
				else {
					message("⤦Ȣ˸ϳݤäƤʤ");
				}
                return 1;
            default:
                break;
            }
        }
        else {
            if (Not(achievements_cabinet_opened) ||
                Is(achievements_treasure_chest_unlocked))
            {
                ReturnNotExist;
            }
            switch (vid){
            case vi_get:
                command_get_item(nid, fi_achievements_wooden_key_taken);
                return 1;
            default:
                break;
            }
        }
		break;
    case ni_sponge:
        if (has_item(nid)){
            switch (vid){
            case vi_use:
				if (Is(achievements_all_completed)){
                    if (Not(achievements_magic_circle_erased)){
        			    message("򤳤⤫Ӿ夬äˡؤä");
    					Set(achievements_magic_circle_erased);
					    lose_item(nid);
                        if (Is(achievements_door_opened)){
                            tentacle_attack();
                        }
                    }
                    else {
                        message("ˡؤϤ⤦äƤ롣");
                    }
                }
                else {
                    message("򤳤Ⱦˤʤä");
                }
                return 1;
            default:
                break;
            }
        }
        else {
            if (Not(achievements_cabinet_opened) ||
                Is(achievements_magic_circle_erased))
            {
                ReturnNotExist;
            }
            switch (vid){
            case vi_get:
                command_get_item(nid, fi_achievements_sponge_taken);
                Set(sponge_wet); // ص塢Ǩ餷Ƥ
                return 1;
            default:
                break;
            }
        }
		break;
    case ni_rubber_duck:
        if (Is(achievements_duck_gone)){
            message("ϼᡢҥϻ");
            return 1;
        }
        else if (Not(achievements_treasure_chest_opened)){
            ReturnNotExist;
        }
        switch (vid){
        case vi_look:
            if (Not(achievements_duck_moved)){
                message("%nnäƤ롣", ni_treasure_chest);
            }
            else {
                message("%nn˾äƤ롣", ni_dog);
            }
            break;
        case vi_move:
            if (Not(achievements_duck_moved)){
                message("ʤ%nn%nn˾褻",
                        ni_rubber_duck, ni_dog);
                Set(achievements_duck_moved);
            }
            else {
                message("ʤ%nn%nnᤷ",
                        ni_rubber_duck, ni_treasure_chest);
                Reset(achievements_duck_moved);
            }
            return 1;
        default:
            break;
        }
        break;
    case ni_mirror:
        if (Not(achievements_death_completed)) ReturnNotExist;  
        switch (vid){
        case vi_look:
            message("ȡޤ路ε롣");
            return 1;
        case vi_examine:
            message("ϻλۼԤζ");
            return 1;
        default:
            break;
        }
        break;
    case ni_window:
        if (Not(achievements_alternative_completed)) ReturnNotExist;
        switch (vid){
        case vi_look:
            message("ˡˡ");
            message("γˤ̾񤤰۳ͻҤ롣");
            return 1;
        case vi_examine:
            message("λۼԤ");
            return 1;
        default:
            break;
        }
        break;
    case ni_darts_target:
        if (Not(achievements_progress_completed)) ReturnNotExist;
        switch (vid){
        case vi_look:
            message("ڤ줿ΰΤˡ"
                    "֥ȥ롼꿩ס֥ƥĸλ"
                    "ʤɤȽ񤭹Ǥ롣"
                    "ȳѡ");
            return 1;
        case vi_examine:
            message("ϿĽλۼԤŪ");
            return 1;
        default:
            break;
        }
        break;
    case ni_plush_cat:
        if (Not(achievements_complete_completed)) ReturnNotExist;
        switch (vid){
        case vi_look:
            message("̤ߤΤդ򤷤Ƥ뤬"
                    "ĤϵۼԤΰ͡%w");
            message("Ĥ⤷ʤ");
            return 1;
        case vi_examine:
            message("ϤꤳߤλۼԤΤ̤ߤ");
            return 1;
        default:
            break;
        }
        break;
    case ni_floor:
        if (Is(achievements_all_completed) &&
            Not(achievements_magic_circle_erased))
        {
            switch (vid){
            case vi_look:
            case vi_examine:
                message("ξˡؤ⤫Ӿ夬äƤ롣");
                return 1;
            default:
                break;
            }
            break;
        }
        break;
    case ni_magic_circle:
        if (Not(achievements_all_completed) ||
            Is(achievements_magic_circle_erased))
        {
            ReturnNotExist;
        }
        switch (vid){
        case vi_look:
            message("󤯤ˡؤ");
            return 1;
        case vi_examine:
            message("Ϥ٤ƤλۼԤˡؤ");
            return 1;
        default:
            break;
        }
        break;
    case ni_chicken:
        if (Not(achievements_treasure_chest_opened)) ReturnNotExist;
        break;
    case ni_ceiling:
    case ni_wall:
        break;
    default:
        *exist = 0;
        // ã٥ƥϤޤȤƽ
        if (command_symbols(vid, nid, exist)) return 1;
        break;
    }
    return 0;
}

// ե饰IDfids(IDinvalidǤʤ)ե饰٤ONʤ0
// OFFΥե饰ĤǤ⤢0֤
int is_achievements_on(flag_id_t fids[]){
    int i;
    for (i = 0 ; i < SYMBOL_FLAG_NUM ; i++){
        //        printf("is_ach_on %d %d %d\n", i, fids[i], is_flag_off(fids[i]));
        if (fids[i] != fi_invalid && is_flag_off(fids[i])) return 0;
    }
    return 1;
}

int count_achievements(int num, symbol_t symbols[]){
    int i, count = 0;
    for (i = 0 ; i < num ; i++){
        //        printf("%s i=%d\n", symbols[i].message, i);
        // invalidǤʤե饰OFFΤΤĤǤ⤢
        // ΥܥϥȤʤ
        if (is_achievements_on(symbols[i].fids)) count ++;
    }
    return count;
}

void message_active_symbols(int num, symbol_t symbols[]){
    int i, count, outcount = 0;
    count = count_achievements(num, symbols);
    switch (count){
    case 0:
        message("ʤ");
        break;
    default:
        for (i = 0 ; i < num ; i++){
            if (is_achievements_on(symbols[i].fids)){
                outcount ++;
                if (outcount < count-1){
                    message("%nl\\", symbols[i].nid);
                }
                else if (outcount == count-1){
                    message("%nl\\", symbols[i].nid);
                }
                else {
                    message("%nl롣", symbols[i].nid);
                }
            }
        }
        break;
    }
}


void message_achievements_count(char *header, int num, symbol_t symbols[]){
    int count;
    char *hint;

    count = count_achievements(num, symbols);
    hint = achievements_hint(num, symbols);
    if (hint == NULL) hint = "";
    message("ʤˤ񤤤Ƥ롣");
    message("%sãΨ %d/%d %s", header, count, num, hint);
}

char *achievements_hint(int num, symbol_t symbols[]){
    int count, ix = 0;

    count = count_achievements(num, symbols);
    if (count == num) return NULL;
    while (ix < num && is_achievements_on(symbols[ix].fids)){
        ix ++;
    }
    assert(ix < num);
    return symbols[ix].hint;
}

flag_id_t *symbol_nid2fids(noun_id_t nid, int num, symbol_t symbols[]){
    int i;
    for (i = 0 ; i < num ; i++){
        if (symbols[i].nid == nid) return symbols[i].fids;
    }
    assert(0);
    return NULL;
}

char *symbol_nid2message(noun_id_t nid, int num, symbol_t symbols[]){
    int i;
    for (i = 0 ; i < num ; i++){
        if (symbols[i].nid == nid) return symbols[i].message;
    }
    assert(0);
    return NULL;
}

void message_symbol(noun_id_t nid, int num, symbol_t symbols[]){
    message("%sפȹޤƤ롣",
            symbol_nid2message(nid, num, symbols));
}

void update_achievements_complete_flags(void){
    if (count_achievements(DEATH_SYMBOL_NUM, death_symbols) ==
        DEATH_SYMBOL_NUM)
    {
        Set(achievements_death_completed);
    }
    else {
        Reset(achievements_death_completed);
    }
    if (count_achievements(PROGRESS_SYMBOL_NUM, progress_symbols) ==
        PROGRESS_SYMBOL_NUM)
    {
        Set(achievements_progress_completed);
    }
    else {
        Reset(achievements_progress_completed);
    }
    if (count_achievements(ALTERNATIVE_SYMBOL_NUM, alternative_symbols) ==
        ALTERNATIVE_SYMBOL_NUM)
    {
        Set(achievements_alternative_completed);
    }
    else {
        Reset(achievements_alternative_completed);
    }
    if (count_achievements(COMPLETE_SYMBOL_NUM, complete_symbols) ==
        COMPLETE_SYMBOL_NUM)
    {
        Set(achievements_complete_completed);
    }
    else {
        Reset(achievements_complete_completed);
    }
    if (count_on_flags(4,
                       fi_achievements_death_completed,
                       fi_achievements_progress_completed,
                       fi_achievements_alternative_completed,
                       fi_achievements_complete_completed) == 4)
    {
        Set(achievements_all_completed);
    }
    else {
        Reset(achievements_all_completed);
    }
}


void tentacle_duck(void){
    Reset(achievements_door_opened);
    Set(achievements_duck_gone);
}

void tentacle_attack(void){
    if (Is(achievements_duck_moved) && Not(achievements_duck_gone)){
        msg_tentacle_duck();
        set_wait_return_callback(tentacle_duck);
    }
    else {
        die_tentacle();
    }
}


int command_symbols(verb_id_t vid, noun_id_t nid, int *exist){
    int i, j;
    symbol_t *symbols;
    int num;
    
    for (i = 0 ; i < SYMBOLS_NUM ; i++){
        symbols = symbol_table[i].symbols;
        num = symbol_table[i].num;
        for (j = 0 ; j < num ; j++){
            if (symbols[j].nid == nid){
                if (!is_achievements_on(symbols[j].fids)){
                    ReturnNotExist;
                }
                switch (vid){
                case vi_look:
                    message(symbols[j].description);
                    return 1;
                case vi_examine:
                    message_symbol(nid, num, symbols);
                    return 1;
                default:
                    break;
                }
                *exist = 1;
                return 0;
            }
        }
    }
    return 0;
}
