#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <stdarg.h>
#include <string.h>
#include "getarg.h"

typedef enum {
    vt_unset = 0, vt_char, vt_int, vt_str
} val_type_t;

typedef struct {
    char *name;
    char *argname;
    char *description;
    int(*func)(int, char *);
    val_type_t val_type;
    union val_u{
        char charval;
        int intval;
        char *strval;
    } val;
} option_t;

#define OPTION_NUM_MAX 64

static option_t options[OPTION_NUM_MAX] = {
    {"help", NULL, "ץɽ", opt_help},
    {"savedir", "dir", "֥ǥ쥯ȥ", opt_set_str},
};
static int option_num = 2;

static char *argv0;

void register_option(char *name, char *argname, char *description, 
                     int(*func)(int, char *)){
    options[option_num].name = name;
    options[option_num].argname = argname;
    options[option_num].description = description;
    options[option_num].func = func;
    option_num ++;
}

void print_current_options(void){
    int i;
    printf("%d option(s) registered.\n", option_num);
    for (i = 0 ; i < option_num ; i++){
        printf("%d %-10s %-6s %s\n", 
               i,
               options[i].name,
               options[i].argname ? options[i].argname : "",
               options[i].description);
        switch(options[i].val_type){
        case vt_unset:
            // ץ̤ʤʤˤ⤷ʤ
            break;
        case vt_char:
            printf("  '%c'\n", options[i].val.charval);
            break;
        case vt_int:
            printf("  %d\n", options[i].val.intval);
            break;
        case vt_str:
            printf("  \"%s\"\n", options[i].val.strval);
            break;
        default:
            assert(0);
            break;
        }
    }
}


int option_name2no(char *name){
    int i;

    for (i = 0 ; i < option_num ; i++){
        if (strcmp(name, options[i].name) == 0) return i;
    }
    return -1;
}


int getarg(int argc, char *argv[]){
    int i, no;

    argv0 = argv[0];
    for (i = 1 ; i < argc ; i++){
        switch (argv[i][0]){
        case '-':
            switch(argv[i][1]){
            case '-':
                no = option_name2no(&argv[i][2]);
                // ̤ΤΥץ
                if (no < 0){
                    opt_help(0, NULL);
                    return 1;
                }
                // ΤΥץ
                else {
                    // դξ
                    if (options[no].argname){
                        if (i+1 < argc){
                            i++;
                            if (options[no].func(no, argv[i])) return 1;
                        }
                        else { // եץΰʤ
                            opt_help(no, NULL);
                            return 1;
                        }
                    }
                    else {
                        if (options[no].func(no, NULL)) return 1;
                    }
                    break;
                }
                break;
            default:
                break;
            }
            break;
        default:
            break;
        }
    }
    // !!! ֥ѥѥλꥪץä顢ʲϽ
    //    set_save_dpath(argv0, save_dir, save_dirname);
    //    set_log_fpath(LOG_FILE);
    return 0;
}

int opt_help(int no, char *arg){
    int i;

    assert(arg == NULL);

    fprintf(stderr, "Ȥ: %s [options]\n", argv0);
    for (i = 0 ; i < option_num ; i++){
        printf("  --%-10s %-4s : %s\n", 
               options[i].name, 
               options[i].argname ? options[i].argname : "",
               options[i].description);
    }
    return -1;
}

int opt_set_char(int no, char *arg){
    printf("set char %d:%c\n", no, arg[0]);
    options[no].val_type = vt_char;
    options[no].val.charval = arg[0];
    return 0;
}
int opt_set_int(int no, char *arg){
    int val;

    val = atoi(arg);
    printf("set int %d:%d\n", no, val);
    options[no].val_type = vt_int;
    options[no].val.intval = val;
    return 0;
}
int opt_set_str(int no, char *arg){
    printf("set str %d:%s\n", no, arg);
    options[no].val_type = vt_str;
    options[no].val.strval = arg;
    return 0;
}

int is_opt_set(char *name){
    int no;
    no = option_name2no(name);
    if (no < 0) return 0;
    return options[no].val_type;
}

char get_opt_char(char *name){
    int no;
    no = option_name2no(name);
    if (no < 0) return 0;
    return options[no].val.charval;
}
int get_opt_int(char *name){
    int no;
    no = option_name2no(name);
    if (no < 0) return 0;
    return options[no].val.intval;
}
char *get_opt_str(char *name){
    int no;
    no = option_name2no(name);
    if (no < 0) return NULL;
    return options[no].val.strval;
}
