#include <stdio.h>
#include <string.h>
#include <assert.h>
#include "roman_conv.h"

typedef enum {
    v_invalid = 0,
    v_l = 1,
    v_x,
    v_a, v_i, v_u, v_e, v_o,
    v_k,
    v_s,
    v_t,
    v_n,
    v_h,
    v_m,
    v_y,
    v_r,
    v_w,

    v_g,
    v_z,
    v_d,
    v_b,
    v_p,

    v_c,
    v_f,
    v_j,
    v_q,
    v_v,
    v_minus,
} char_value_t;

#define IsVowel(v) ((v)<=v_o)

// char_value­碌Ȥ˥եȤӥåȿ
// char_value_tκͤǼޤΥӥåȿɬ
#define SHIFT 5
// 
#define STR_VALUE_MAX ((v_l<<(SHIFT*3))+(v_t<<(SHIFT*2))+(v_s<<SHIFT)+v_u)


typedef struct {
    unsigned char c;
    char_value_t cv;
} c_cv_t;
c_cv_t c_cv[] = {
    {'\0', v_invalid},
    {' ', v_invalid},
    {'a', v_a},
    {'i', v_i},
    {'u', v_u},
    {'e', v_e},
    {'o', v_o},
    {'k', v_k},
    {'s', v_s},
    {'t', v_t},
    {'n', v_n},
    {'h', v_h},
    {'m', v_m},
    {'y', v_y},
    {'r', v_r},
    {'w', v_w},

    {'g', v_g},
    {'z', v_z},
    {'d', v_d},
    {'b', v_b},
    {'p', v_p},

    {'c', v_c},
    {'f', v_f},
    {'j', v_j},
    {'l', v_l},
    {'q', v_q},
    {'v', v_v},
    {'-', v_minus},
};

typedef struct {
    char *hiragana;
    char *katakana;
    // ʤɡҤ餬ʤȥʤĹۤʤ礬
    int hira_len;
    int kata_len;
} kana_t;

typedef int str_value_t;

typedef struct {
    str_value_t sv;
    kana_t kana;
} sv_kana_t;

#define SV_NN ((v_n<<SHIFT)+v_n)
#define SV_LTU ((v_l<<(SHIFT*2))+(v_t<<SHIFT)+v_u)

sv_kana_t sv_kana[] = {
    {v_a,      {"", "", 0}},
    {v_i,      {"", "", 0}},
    {v_u,      {"", "", 0}},
    {v_e,      {"", "", 0}},
    {v_o,      {"", "", 0}},
    {(v_k<<SHIFT)+v_a,  {"", "", 0}},
    {(v_k<<SHIFT)+v_i,  {"", "", 0}},
    {(v_k<<SHIFT)+v_u,  {"", "", 0}},
    {(v_k<<SHIFT)+v_e,  {"", "", 0}},
    {(v_k<<SHIFT)+v_o,  {"", "", 0}},
    {(v_s<<SHIFT)+v_a,  {"", "", 0}},
    {(v_s<<SHIFT)+v_i,  {"", "", 0}},
    {(v_s<<SHIFT)+v_u,  {"", "", 0}},
    {(v_s<<SHIFT)+v_e,  {"", "", 0}},
    {(v_s<<SHIFT)+v_o,  {"", "", 0}},
    {(v_t<<SHIFT)+v_a,  {"", "", 0}},
    {(v_t<<SHIFT)+v_i,  {"", "", 0}},
    {(v_t<<SHIFT)+v_u,  {"", "", 0}},
    {(v_t<<SHIFT)+v_e,  {"", "", 0}},
    {(v_t<<SHIFT)+v_o,  {"", "", 0}},
    {(v_n<<SHIFT)+v_a,  {"", "", 0}},
    {(v_n<<SHIFT)+v_i,  {"", "", 0}},
    {(v_n<<SHIFT)+v_u,  {"", "", 0}},
    {(v_n<<SHIFT)+v_e,  {"", "", 0}},
    {(v_n<<SHIFT)+v_o,  {"", "", 0}},
    {(v_h<<SHIFT)+v_a,  {"", "", 0}},
    {(v_h<<SHIFT)+v_i,  {"", "", 0}},
    {(v_h<<SHIFT)+v_u,  {"", "", 0}},
    {(v_h<<SHIFT)+v_e,  {"", "", 0}},
    {(v_h<<SHIFT)+v_o,  {"", "", 0}},
    {(v_m<<SHIFT)+v_a,  {"", "", 0}},
    {(v_m<<SHIFT)+v_i,  {"", "", 0}},
    {(v_m<<SHIFT)+v_u,  {"", "", 0}},
    {(v_m<<SHIFT)+v_e,  {"", "", 0}},
    {(v_m<<SHIFT)+v_o,  {"", "", 0}},
    {(v_y<<SHIFT)+v_a,  {"", "", 0}},
    {(v_y<<SHIFT)+v_u,  {"", "", 0}},
    {(v_y<<SHIFT)+v_o,  {"", "", 0}},
    {(v_r<<SHIFT)+v_a,  {"", "", 0}},
    {(v_r<<SHIFT)+v_i,  {"", "", 0}},
    {(v_r<<SHIFT)+v_u,  {"", "", 0}},
    {(v_r<<SHIFT)+v_e,  {"", "", 0}},
    {(v_r<<SHIFT)+v_o,  {"", "", 0}},
    {(v_w<<SHIFT)+v_a,  {"", "", 0}},
    {(v_w<<SHIFT)+v_i,  {"", "", 0}},
    {(v_w<<SHIFT)+v_u,  {"", "", 0}},
    {(v_w<<SHIFT)+v_e,  {"", "", 0}},
    {(v_w<<SHIFT)+v_o,  {"", "", 0}},
    {(v_g<<SHIFT)+v_a,  {"", "", 0}},
    {(v_g<<SHIFT)+v_i,  {"", "", 0}},
    {(v_g<<SHIFT)+v_u,  {"", "", 0}},
    {(v_g<<SHIFT)+v_e,  {"", "", 0}},
    {(v_g<<SHIFT)+v_o,  {"", "", 0}},
    {(v_z<<SHIFT)+v_a,  {"", "", 0}},
    {(v_z<<SHIFT)+v_i,  {"", "", 0}},
    {(v_z<<SHIFT)+v_u,  {"", "", 0}},
    {(v_z<<SHIFT)+v_e,  {"", "", 0}},
    {(v_z<<SHIFT)+v_o,  {"", "", 0}},
    {(v_d<<SHIFT)+v_a,  {"", "", 0}},
    {(v_d<<SHIFT)+v_i,  {"", "", 0}},
    {(v_d<<SHIFT)+v_u,  {"", "", 0}},
    {(v_d<<SHIFT)+v_e,  {"", "", 0}},
    {(v_d<<SHIFT)+v_o,  {"", "", 0}},
    {(v_b<<SHIFT)+v_a,  {"", "", 0}},
    {(v_b<<SHIFT)+v_i,  {"", "", 0}},
    {(v_b<<SHIFT)+v_u,  {"", "", 0}},
    {(v_b<<SHIFT)+v_e,  {"", "", 0}},
    {(v_b<<SHIFT)+v_o,  {"", "", 0}},
    {(v_p<<SHIFT)+v_a,  {"", "", 0}},
    {(v_p<<SHIFT)+v_i,  {"", "", 0}},
    {(v_p<<SHIFT)+v_u,  {"", "", 0}},
    {(v_p<<SHIFT)+v_e,  {"", "", 0}},
    {(v_p<<SHIFT)+v_o,  {"", "", 0}},

    {SV_NN,  {"", "", 0}},
    {v_minus,      {"", "", 0}},

    {(v_f<<SHIFT)+v_a,  {"դ", "ե", 0}},
    {(v_f<<SHIFT)+v_i,  {"դ", "ե", 0}},
    {(v_f<<SHIFT)+v_u,  {"", "", 0}},
    {(v_f<<SHIFT)+v_e,  {"դ", "ե", 0}},
    {(v_f<<SHIFT)+v_o,  {"դ", "ե", 0}},
    {(v_j<<SHIFT)+v_a,  {"", "", 0}},
    {(v_j<<SHIFT)+v_i,  {"", "", 0}},
    {(v_j<<SHIFT)+v_u,  {"", "", 0}},
    {(v_j<<SHIFT)+v_e,  {"", "", 0}},
    {(v_j<<SHIFT)+v_o,  {"", "", 0}},
    {(v_l<<SHIFT)+v_a,  {"", "", 0}},
    {(v_l<<SHIFT)+v_i,  {"", "", 0}},
    {(v_l<<SHIFT)+v_u,  {"", "", 0}},
    {(v_l<<SHIFT)+v_e,  {"", "", 0}},
    {(v_l<<SHIFT)+v_o,  {"", "", 0}},
    {(v_x<<SHIFT)+v_a,  {"", "", 0}},
    {(v_x<<SHIFT)+v_i,  {"", "", 0}},
    {(v_x<<SHIFT)+v_u,  {"", "", 0}},
    {(v_x<<SHIFT)+v_e,  {"", "", 0}},
    {(v_x<<SHIFT)+v_o,  {"", "", 0}},
    {(v_v<<SHIFT)+v_a,  {"", "", 0}},
    {(v_v<<SHIFT)+v_i,  {"", "", 0}},
    {(v_v<<SHIFT)+v_u,  {"", "", 0}},
    {(v_v<<SHIFT)+v_e,  {"", "", 0}},
    {(v_v<<SHIFT)+v_o,  {"", "", 0}},
    {(v_q<<SHIFT)+v_a,  {"", "", 0}},
    {(v_q<<SHIFT)+v_i,  {"", "", 0}},
    {(v_q<<SHIFT)+v_u,  {"", "", 0}},
    {(v_q<<SHIFT)+v_e,  {"", "", 0}},
    {(v_q<<SHIFT)+v_o,  {"", "", 0}},

    {(v_b<<(SHIFT*2))+(v_y<<SHIFT)+v_a,  {"Ӥ", "ӥ", 0}},
    {(v_b<<(SHIFT*2))+(v_y<<SHIFT)+v_i,  {"Ӥ", "ӥ", 0}},
    {(v_b<<(SHIFT*2))+(v_y<<SHIFT)+v_u,  {"Ӥ", "ӥ", 0}},
    {(v_b<<(SHIFT*2))+(v_y<<SHIFT)+v_e,  {"Ӥ", "ӥ", 0}},
    {(v_b<<(SHIFT*2))+(v_y<<SHIFT)+v_o,  {"Ӥ", "ӥ", 0}},
    {(v_t<<(SHIFT*2))+(v_h<<SHIFT)+v_a,  {"Ƥ", "ƥ", 0}},
    {(v_t<<(SHIFT*2))+(v_h<<SHIFT)+v_i,  {"Ƥ", "ƥ", 0}},
    {(v_t<<(SHIFT*2))+(v_h<<SHIFT)+v_u,  {"Ƥ", "ƥ", 0}},
    {(v_t<<(SHIFT*2))+(v_h<<SHIFT)+v_e,  {"Ƥ", "ƥ", 0}},
    {(v_t<<(SHIFT*2))+(v_h<<SHIFT)+v_o,  {"Ƥ", "ƥ", 0}},
    {(v_t<<(SHIFT*2))+(v_w<<SHIFT)+v_u,  {"Ȥ", "ȥ", 0}},
    {(v_d<<(SHIFT*2))+(v_h<<SHIFT)+v_a,  {"Ǥ", "ǥ", 0}},
    {(v_d<<(SHIFT*2))+(v_h<<SHIFT)+v_i,  {"Ǥ", "ǥ", 0}},
    {(v_d<<(SHIFT*2))+(v_h<<SHIFT)+v_u,  {"Ǥ", "ǥ", 0}},
    {(v_d<<(SHIFT*2))+(v_h<<SHIFT)+v_e,  {"Ǥ", "ǥ", 0}},
    {(v_d<<(SHIFT*2))+(v_h<<SHIFT)+v_o,  {"Ǥ", "ǥ", 0}},
    {(v_d<<(SHIFT*2))+(v_w<<SHIFT)+v_u,  {"ɤ", "ɥ", 0}},
    {(v_c<<(SHIFT*2))+(v_h<<SHIFT)+v_a,  {"", "", 0}},
    {(v_c<<(SHIFT*2))+(v_h<<SHIFT)+v_i,  {"", "", 0}},
    {(v_c<<(SHIFT*2))+(v_h<<SHIFT)+v_u,  {"", "", 0}},
    {(v_c<<(SHIFT*2))+(v_h<<SHIFT)+v_e,  {"", "", 0}},
    {(v_c<<(SHIFT*2))+(v_h<<SHIFT)+v_o,  {"", "", 0}},
    {(v_h<<(SHIFT*2))+(v_y<<SHIFT)+v_a,  {"Ҥ", "ҥ", 0}},
    {(v_h<<(SHIFT*2))+(v_y<<SHIFT)+v_i,  {"Ҥ", "ҥ", 0}},
    {(v_h<<(SHIFT*2))+(v_y<<SHIFT)+v_u,  {"Ҥ", "ҥ", 0}},
    {(v_h<<(SHIFT*2))+(v_y<<SHIFT)+v_e,  {"Ҥ", "ҥ", 0}},
    {(v_h<<(SHIFT*2))+(v_y<<SHIFT)+v_o,  {"Ҥ", "ҥ", 0}},
    {(v_s<<(SHIFT*2))+(v_h<<SHIFT)+v_a,  {"", "", 0}},
    {(v_s<<(SHIFT*2))+(v_h<<SHIFT)+v_i,  {"", "", 0}},
    {(v_s<<(SHIFT*2))+(v_h<<SHIFT)+v_u,  {"", "", 0}},
    {(v_s<<(SHIFT*2))+(v_h<<SHIFT)+v_e,  {"", "", 0}},
    {(v_s<<(SHIFT*2))+(v_h<<SHIFT)+v_o,  {"", "", 0}},
    {(v_s<<(SHIFT*2))+(v_y<<SHIFT)+v_a,  {"", "", 0}},
    {(v_s<<(SHIFT*2))+(v_y<<SHIFT)+v_i,  {"", "", 0}},
    {(v_s<<(SHIFT*2))+(v_y<<SHIFT)+v_u,  {"", "", 0}},
    {(v_s<<(SHIFT*2))+(v_y<<SHIFT)+v_e,  {"", "", 0}},
    {(v_s<<(SHIFT*2))+(v_y<<SHIFT)+v_o,  {"", "", 0}},
    {(v_k<<(SHIFT*2))+(v_y<<SHIFT)+v_a,  {"", "", 0}},
    {(v_k<<(SHIFT*2))+(v_y<<SHIFT)+v_i,  {"", "", 0}},
    {(v_k<<(SHIFT*2))+(v_y<<SHIFT)+v_u,  {"", "", 0}},
    {(v_k<<(SHIFT*2))+(v_y<<SHIFT)+v_e,  {"", "", 0}},
    {(v_k<<(SHIFT*2))+(v_y<<SHIFT)+v_o,  {"", "", 0}},
    {(v_g<<(SHIFT*2))+(v_y<<SHIFT)+v_a,  {"", "", 0}},
    {(v_g<<(SHIFT*2))+(v_y<<SHIFT)+v_i,  {"", "", 0}},
    {(v_g<<(SHIFT*2))+(v_y<<SHIFT)+v_u,  {"", "", 0}},
    {(v_g<<(SHIFT*2))+(v_y<<SHIFT)+v_e,  {"", "", 0}},
    {(v_g<<(SHIFT*2))+(v_y<<SHIFT)+v_o,  {"", "", 0}},

    {(v_m<<(SHIFT*2))+(v_y<<SHIFT)+v_a,  {"ߤ", "ߥ", 0}},
    {(v_m<<(SHIFT*2))+(v_y<<SHIFT)+v_i,  {"ߤ", "ߥ", 0}},
    {(v_m<<(SHIFT*2))+(v_y<<SHIFT)+v_u,  {"ߤ", "ߥ", 0}},
    {(v_m<<(SHIFT*2))+(v_y<<SHIFT)+v_e,  {"ߤ", "ߥ", 0}},
    {(v_m<<(SHIFT*2))+(v_y<<SHIFT)+v_o,  {"ߤ", "ߥ", 0}},
    {(v_n<<(SHIFT*2))+(v_y<<SHIFT)+v_a,  {"ˤ", "˥", 0}},
    {(v_n<<(SHIFT*2))+(v_y<<SHIFT)+v_i,  {"ˤ", "˥", 0}},
    {(v_n<<(SHIFT*2))+(v_y<<SHIFT)+v_u,  {"ˤ", "˥", 0}},
    {(v_n<<(SHIFT*2))+(v_y<<SHIFT)+v_e,  {"ˤ", "˥", 0}},
    {(v_n<<(SHIFT*2))+(v_y<<SHIFT)+v_o,  {"ˤ", "˥", 0}},

    {(v_r<<(SHIFT*2))+(v_y<<SHIFT)+v_a,  {"", "", 0}},
    {(v_r<<(SHIFT*2))+(v_y<<SHIFT)+v_i,  {"ꤣ", "ꥣ", 0}},
    {(v_r<<(SHIFT*2))+(v_y<<SHIFT)+v_u,  {"", "", 0}},
    {(v_r<<(SHIFT*2))+(v_y<<SHIFT)+v_e,  {"ꤧ", "ꥧ", 0}},
    {(v_r<<(SHIFT*2))+(v_y<<SHIFT)+v_o,  {"", "", 0}},

    {(v_q<<(SHIFT*2))+(v_w<<SHIFT)+v_a,  {"", "", 0}},
    {(v_q<<(SHIFT*2))+(v_w<<SHIFT)+v_i,  {"", "", 0}},
    {(v_q<<(SHIFT*2))+(v_w<<SHIFT)+v_u,  {"", "", 0}},
    {(v_q<<(SHIFT*2))+(v_w<<SHIFT)+v_e,  {"", "", 0}},
    {(v_q<<(SHIFT*2))+(v_w<<SHIFT)+v_o,  {"", "", 0}},

    {(v_t<<(SHIFT*2))+(v_s<<SHIFT)+v_a,  {"Ĥ", "ĥ", 0}},
    {(v_t<<(SHIFT*2))+(v_s<<SHIFT)+v_i,  {"Ĥ", "ĥ", 0}},
    {(v_t<<(SHIFT*2))+(v_s<<SHIFT)+v_u,  {"",   "",   0}},
    {(v_t<<(SHIFT*2))+(v_s<<SHIFT)+v_e,  {"Ĥ", "ĥ", 0}},
    {(v_t<<(SHIFT*2))+(v_s<<SHIFT)+v_o,  {"Ĥ", "ĥ", 0}},

    {(v_l<<(SHIFT*2))+(v_y<<SHIFT)+v_a,  {"", "", 0}},
    {(v_l<<(SHIFT*2))+(v_y<<SHIFT)+v_u,  {"", "", 0}},
    {(v_l<<(SHIFT*2))+(v_y<<SHIFT)+v_o,  {"", "", 0}},
    {SV_LTU,  {"", "", 0}},
    //    {(v_l<<(SHIFT*2))+(v_t<<SHIFT)+v_u,  {"", "", 0}},
    {(v_l<<(SHIFT*3))+(v_t<<(SHIFT*2))+(v_s<<SHIFT)+v_u,  {"", "", 0}},
};

static char_value_t c2cv[256];
static kana_t *sv2kana[STR_VALUE_MAX+1];

static roman_conv_mode_t kanamode = rc_mode_hiragana;

void put_kana(char **kstr, kana_t *kana);
static char is_initialized = 0;


int roman_conv_init(void){
	int i;

    if (is_initialized) return 1;

    for (i = 0 ; i < sizeof(c_cv)/sizeof(c_cv_t) ; i++){
        assert(c2cv[c_cv[i].c] == 0);
        //                printf("c_cv %c %d\n", c_cv[i].c, c_cv[i].cv);
        c2cv[c_cv[i].c] = c_cv[i].cv;
    }
    for (i = 0 ; i < sizeof(sv_kana)/sizeof(sv_kana_t) ; i++){
        //        printf("sv_k %d %s\n", sv_kana[i].sv, sv_kana[i].kana.hiragana);
        sv_kana[i].kana.hira_len = strlen(sv_kana[i].kana.hiragana);
        sv_kana[i].kana.kata_len = strlen(sv_kana[i].kana.katakana);
        sv2kana[sv_kana[i].sv] = &sv_kana[i].kana;
    }
    is_initialized = 1;
    return 0;
}

void roman_conv_set_mode(roman_conv_mode_t mode){
    assert(mode == rc_mode_hiragana ||
           mode == rc_mode_katakana);
    kanamode = mode;
}

// rstrƬ޻ѴԤä̤ʸkstr˳Ǽ
// ѴǤʤäȤϡĤäե٥å(1Хʸ)ʬ
// Ƭ֤leftstr˳Ǽ
// ѴǤȤ0, ʳʤ0֤
// terminated0ΤȤϡϤλƤ(Ȥnʤ֤פˤ)
// terminated0ΤȤϡϤԴ(Ȥnʤ֤פѴʤ)
int roman2kana(char* rstr, char *kstr, char **leftstr, int terminated){
    int cv, cv_next;
    int sv = 0;
	kana_t *kana;

    *kstr = '\0';

    while (*rstr != '\0'){
        cv = c2cv[(unsigned char)*rstr];
        if (cv == v_invalid) break;

        cv_next = c2cv[(unsigned char)*(rstr+1)];
        
        sv = (sv << SHIFT) + cv;
        // λ'nn'ˤʤäƤ餺'n'θ³ʸ'n'ʳλҲʤ顢
        // 'nn'ɤȤˤ''Ѵ
        if (cv == v_n && sv != SV_NN){
            if ((cv_next != v_n && cv_next != v_y && !IsVowel(cv_next)) ||
                (cv_next == v_invalid && terminated != 0))
            {
                sv = (sv << SHIFT) + cv;
            }
        }
        // nnξҲϢ³
        // ǽλҲ(cv)''Ѵ
        else if (cv == cv_next && !IsVowel(cv) && sv != SV_NN){ 
            sv = SV_LTU;
        }
        if (sv > STR_VALUE_MAX) return -1;
        kana = sv2kana[sv];
        rstr ++;
        if (kana != NULL){
            put_kana(&kstr, kana);
            sv = 0;
            *leftstr = rstr;
        }
    }
    if (*leftstr == rstr){
        return 0;
    }
    else {
        return -1;
    }
}

void put_kana(char **kstr, kana_t *kana){
    assert(kana != NULL);

    if (kanamode == rc_mode_hiragana){
        strcpy(*kstr, kana->hiragana);
        *kstr += kana->hira_len;
    }
    else {
        strcpy(*kstr, kana->katakana);
        *kstr += kana->kata_len;
    }
    
}
