#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <EzGraph.h>
#include "config.h"
#include "layout.h"
#include "term.h"
#include "command.h"
#include "history.h"
#include "screen.h"
#include "graphic.h"
#include "map.h"
#include "console.h"
#include "roman_input_line.h"
#include "file.h"

//static char input[JBufSize(CONSOLE_CHR_WIDTH)+1];

typedef struct {
    int is_waited;
    char message[MESSAGE_MAX_LEN];
} message_queue_t;

static message_queue_t message_queue[MESSAGE_QUEUE_MAX_LEN];
static int message_queue_tail;
static int message_queue_head;

static char *prompt = DEFAULT_PROMPT;

static void (*wait_return_callback)(void) = NULL;

void draw_console(void);
void erase_input_line(int pos);
void draw_input_line(void);
void term_keyh(int key);
void wait_return_keyh(int key);
void set_term_hist(int n);
void puts_term_waited_callback(void);

int term_init(int lines, int width, int height){
	if (console_init(CONSOLE_BUFFER_LINES,
                     CONSOLE_CHR_WIDTH, CONSOLE_CHR_HEIGHT-1))
    {
        log_error("console_init failed.\n");
        exit(1);
    }
    if (roman_input_line_init(CONSOLE_CHR_WIDTH)){
        log_error("roman_input_line_init failed.\n");
        exit(1);
    }
    roman_input_line_set_cursor(CURSOR);
    set_term_input_mode(DEFAULT_INPUT_MODE);

	EzSetKeyHandler(term_keyh);
    //	EzSetKeyReleaseHandler(term_key_releaseh);
    clear_input();
    return 0;
}

void draw_console(void){
    int i;
    char *line;

    EzSetColorByRGB(0, 0, 0);
    //    EzSetColorByRGB(255, 0, 0);
    EzFillBox(CONSOLE_LT_X-1, CONSOLE_LT_Y-1,
			  CONSOLE_RB_X+1, CONSOLE_RB_Y+1);
    EzSetColorByRGB(255, 255, 255);
    EzDrawBox(CONSOLE_LT_X-2, CONSOLE_LT_Y-2,
		      CONSOLE_RB_X+2, CONSOLE_RB_Y+2);
//	EzDrawBox(CONSOLE_LT_X, CONSOLE_LT_Y,
//		      CONSOLE_LT_X+CONSOLE_WIDTH*FONT_SIZE,
//			  CONSOLE_LT_Y+CONSOLE_HEIGHT*FONT_SIZE);
    console_reset_getline();
    i = 0;
    while ((line = console_getline()) != NULL){
        //        printf("CON %2d:%s\n", i, line);
        EzDrawString(CONSOLE_LT_X, CONSOLE_LT_Y+i*LINE_SKIP+LINE_SKIP, line);
        i++;
    }
}
void erase_input_line(int pos){
    //    EzSetColorByRGB(0, 0, 80+pos*32);
    EzSetColorByRGB(0, 0, 0);
    EzFillBox(CONSOLE_LT_X-1+FONT_SIZE/2*(pos),
              CONSOLE_LT_Y-2+CONSOLE_CHR_HEIGHT*(LINE_SKIP-2),
			  CONSOLE_RB_X+1, CONSOLE_RB_Y+1);
    // @note ɤĤ֤ΰβդRB_Y+1ʤȡ
    // 'j'ʤɤΥեȤǲ1ɥåȤäʤ
}
void draw_input_line(void){
    EzSetColorByRGB(255, 255, 255);
	EzDrawString(CONSOLE_LT_X, CONSOLE_LT_Y+CONSOLE_CHR_HEIGHT*LINE_SKIP, 
                 roman_input_line_get_line());
    //    printf("INPUT|%s|\n", roman_input_line_get_line());
}

void draw_term(void){
    draw_console();
    draw_input_line();
    //    show_console_buf();

}

// puts_term_waited줿å򥭥塼/ǥ塼
// 塼puts_term_waitedϤ줿ΤΤۤ塼ǤʤȤ
// puts/print줿Τ롣
// ǥ塼Ϥޤȡ塼ˤʤޤǥ塼ϹԤʤ

void queue_message(char *line, int is_waited){
    assert(message_queue_tail <= MESSAGE_QUEUE_MAX_LEN);
    strcpy(message_queue[message_queue_tail].message, line);
    message_queue[message_queue_tail].is_waited = is_waited;
    message_queue_tail ++;
}
char *dequeue_message(int *is_waited){
    if (message_queue_head == message_queue_tail){
        message_queue_head = message_queue_tail = 0;
        return NULL;
    }
    *is_waited = message_queue[message_queue_head].is_waited;
    //    *is_wait = (message_queue_head <= message_queue_last_wait) ? 1 : 0;
    return message_queue[message_queue_head ++].message;
}
#define LastQueuedMessage (message_queue[message_queue_tail-1].message)
#define IsMessageQueueEmpty() (message_queue_tail == 0)
#ifdef DEBUG
void print_message_queue(void){
    int i;
    printf("message queue head:%d-tail:%d\n", message_queue_head, message_queue_tail);
    for (i = 0 ; i < message_queue_tail ; i++){
        printf("%d:(%d)%s\n", i, message_queue[i].is_waited, message_queue[i].message);
    }
}
#endif

// @todo
// ߤΤȤprint_termputs_term_waitedȤκߤбƤʤ
void print_term(char *line){
    printf("print :%s:\n", line);
    print_console(line);
    //    draw_term();
}
void puts_term(char *line){
    //    printf("puts :%s:\n", line);
    if (IsMessageQueueEmpty()){
        puts_console(line);
    }
    else {
        queue_message(line, 0);
#ifdef DEBUG
        print_message_queue();
#endif
    }
    draw_term();
}

void puts_term_waited(char *line){
    int is_wait;

    if (IsMessageQueueEmpty()){
        // ǽΥդåϤξǽϤ뤬
        // 2ܰʹߤǤ뤳Ȥ狼褦ˤä󥭥塼
        queue_message(line, 1);
        set_wait_return_callback(puts_term_waited_callback);
        
        puts_console(dequeue_message(&is_wait));
    }
    else {
        queue_message(line, 1);
    }
#ifdef DEBUG
    print_message_queue();
#endif
}

void puts_term_waited_callback(void){
    int is_wait;
    char *message;

    while ((message = dequeue_message(&is_wait)) != NULL){
        puts_console(message);
        if (is_wait){
            set_wait_return_callback(puts_term_waited_callback);        
            return;
        }
    }
}





void clear_input(void){
    //    char buf[256];
    //    roman_input_line_get_input(buf);
    //    printf("hoge:%s:%s\n", roman_input_line_get_line(), buf);
    roman_input_line_clear();
    //    roman_input_line_set_str(CURSOR);
    //    roman_input_line_set_str("_");
    //    roman_input_line_get_input(buf);
    //    printf("hoge:%s:%s\n", roman_input_line_get_line(), buf);
}

void set_term_prompt(char *new_prompt){
    if (new_prompt == NULL){
        new_prompt = DEFAULT_PROMPT;
    }
    prompt = new_prompt;
    roman_input_line_set_prompt(prompt);
}

input_mode_t get_term_input_mode(void){
    return roman_input_line_get_mode();
}
void set_term_input_mode(input_mode_t mode){
    roman_input_line_set_mode(mode);
}

void term_keyh(int key){
    char buf[CONSOLE_CHR_WIDTH];
    int prev_roman_byte, prev_width;
    
	switch (key) {
	case EZ_Page_Up:
	case 'u'-'a'+1:
        console_scroll_up(1);
        draw_term();
		break;
	case EZ_Page_Down:
	case 'd'-'a'+1:
        console_scroll_down(1);
        draw_term();
		break;
	case 8: // BS
        roman_input_line_backspace();
        erase_input_line(roman_input_line_get_width());
        draw_input_line();
        break;
 	case 'c'-'a'+1:
        console_reset_scroll();
        roman_input_line_append_str("^C");
        roman_input_line_get_input(buf);
        print_term(prompt);
        puts_term(buf);
        clear_input();
        draw_term();
		break;
	case EZ_Up:
	case 'p'-'a'+1:
        set_term_hist(-1);
		erase_input_line(0);
        draw_input_line();
        break;
	case EZ_Down:
	case 'n'-'a'+1:
        set_term_hist(1);
		erase_input_line(0);
        draw_input_line();
        break;
	case 13:
        console_reset_scroll();
        roman_input_line_get_input(buf);
        print_term(prompt);
        puts_term(buf);
        interpret_line(buf);
        //		if (rollback > 0){
        //			rollback = 0;
        //		}
        if (wait_return_callback == NULL){
            clear_input();
        }
        draw_term();
		break;
	case EZ_F1:
		set_term_input_mode((get_term_input_mode() +1) % 3);
		printf("change input mode to %d\n", roman_input_line_get_mode());
		break;
	case EZ_F2:
        toggle_show_minimap();
        draw_screen();
        //        draw_graphic();
        break;
	case EZ_F3:
        toggle_minimap_location();
        draw_screen();
        //        draw_graphic();
        break;
	case EZ_F4:
        toggle_overlay_minimap();
        draw_screen();
        //        draw_graphic();
        break;
    default:
        //        printf("KEY:'%c'\n", key);
        // Windowsctype0-255ʳͤassert顼ˤʤؤн
        // ɳü쥳ɤ̵
        // (ɤΤߤ󥽡ϤǤʤʸʤΤ)
//        printf("key:%x(%c) CTRL:%d\n", key, key, control_down);
        if (key & 0xffffff00) return;
        key = tolower(key);
        if (isalnum(key) || key == ' ' || key == '-'){
            //            printf("key:%c\n", key);
            prev_width = roman_input_line_get_width();
            prev_roman_byte = roman_input_line_get_roman_byte();
            roman_input_line_append_chr(key);
            if (roman_input_line_get_roman_byte() == 0){
                //                printf("roman byte:%d\n", prev_roman_byte);
                erase_input_line(prev_width-prev_roman_byte);
            }
            else {
                erase_input_line(prev_width);
            }
#ifdef DEBUG
            roman_input_line_print_line();
#endif
            draw_input_line();
        }
        break;
    }
}


void set_term_hist(int n){
    char *hist;

    if ((roman_input_line_get_word_num() == 1 &&
         roman_input_line_is_tail_spc()) ||
        roman_input_line_get_word_num() == 2)
    {
        set_history_mode(hm_noun);
        while (roman_input_line_get_word_num() > 1){
            roman_input_line_delete_word();
        }
    }
    else {
        set_history_mode(hm_verb);
        clear_input();
    }
    if (n < 0){
		prev_history_line();
    }
    else {
		next_history_line();
    }
    hist = get_history_line();
    if (hist != NULL){
        roman_input_line_append_str(hist);
    }
}

void set_wait_return_callback(void (*callback)(void)){
    printf("set_wait_return_callback called\n");
	EzSetKeyHandler(wait_return_keyh);
    roman_input_line_set_prompt("");
    roman_input_line_set_cursor(' ');
    roman_input_line_set_str("[꥿󥭡򲡤Ƥ]");
#ifdef DEBUG
    roman_input_line_print_line();
#endif
    wait_return_callback = callback;
}

void wait_return_keyh(int key){
    void (*callback)(void);
    printf("wait return keyh\n");
    // callback==NULLʤ餳ΥϥɥϸƤФʤϤ
    assert(wait_return_callback != NULL);
	if (key == 13){
        EzSetKeyHandler(term_keyh);
        callback = wait_return_callback;
        wait_return_callback = NULL;
        roman_input_line_set_prompt(DEFAULT_PROMPT);
        roman_input_line_set_cursor(CURSOR);
        //        set_term_prompt(DEFAULT_PROMPT);
        clear_input();
        puts_console(" ");
        callback();
    }
	// Хåؿǥե饰ѹƤ
	// եåѲ뤫⤷ʤΤǺ
	draw_graphic();
    draw_term();
}






