#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "config.h"
#include "verb.h"
#include "layout.h"
#include "message.h"
#include "status.h"
#include "history.h"
#include "term.h"

static verb_t verbs[] = {
  {vi_invalid, {"", "", ""}, ""},
  {vi_forward, {"forward", "", ""}, "f"},
  {vi_backward, {"backward", "ɤ", "ɥ"}, "b"},
  {vi_left, {"left", "Ҥ", "ҥ"}, "l"},
  {vi_right, {"right", "ߤ", "ߥ"}, "r"},
  {vi_up, {"up", "", ""}, "u"},
  {vi_down, {"down", "", ""}, "d"},
  {vi_enter, {"enter", "Ϥ", "ϥ"}, NULL},
  {vi_exit, {"exit", "Ǥ", "ǥ"}, NULL},
  {vi_inventory, {"inventory", "", ""}, "i"},
  {vi_look, {"look", "ߤ", "ߥ"}, NULL},
  {vi_examine, {"examine", "٤", "٥"}, NULL},
  {vi_recall, {"recall", "⤤", "⥤"}, NULL},
  {vi_get, {"get", "Ȥ", "ȥ"}, NULL},
  {vi_put, {"put", "", ""}, NULL},
  {vi_use, {"use", "Ĥ", "ĥ"}, NULL},
  {vi_open, {"open", "Ҥ餯", "ҥ饯"}, NULL},
  {vi_close, {"close", "Ȥ", "ȥ"}, NULL},
  {vi_hit, {"hit", "", ""}, NULL},
  {vi_kick, {"kick", "", ""}, NULL},
  {vi_fight, {"fight", "", ""}, NULL},
  {vi_destroy, {"break", "魯", "凉"}, NULL},
  {vi_move, {"move", "", ""}, NULL},

  {vi_remove, {"remove", "ɤ", "ɥ"}, NULL},
  {vi_push, {"push", "", ""}, NULL},
  {vi_pull, {"pull", "Ҥ", "ҥ"}, NULL},
  {vi_touch, {"touch", "", ""}, NULL},
  {vi_turn, {"turn", "ޤ魯", "ޥ凉"}, NULL},
  {vi_read, {"read", "", ""}, NULL},
  {vi_throw, {"throw", "ʤ", "ʥ"}, NULL},
  {vi_combine, {"combine", "ä", "å"}, NULL},
  {vi_dismantle, {"dismantle", "֤󤫤", "֥󥫥"}, NULL},
  {vi_adjust, {"adjust", "碌", "糧"}, NULL},
  {vi_pray, {"pray", "Τ", "Υ"}, NULL},
  {vi_talk, {"talk", "Ϥʤ", "ϥʥ"}, NULL},
  {vi_cast, {"cast", "Ȥʤ", "ȥʥ"}, NULL},
  {vi_eat, {"eat", "٤", "٥"}, NULL},
  {vi_drink, {"drink", "Τ", "Υ"}, NULL},
  {vi_wait, {"wait", "ޤ", "ޥ"}, NULL},
  {vi_shoot, {"shoot", "", ""}, NULL},
  {vi_erase, {"erase", "", ""}, NULL},
  {vi_give, {"give", "", ""}, NULL},
  {vi_show, {"show", "ߤ", "ߥ"}, NULL},
  {vi_help, {"help", "ؤ", "إ"}, NULL},
  {vi_load, {"load", "", ""}, NULL},
  {vi_save, {"save", "", ""}, NULL},
  {vi_config, {"config", "äƤ", "åƥ"}, NULL},
  {vi_quit, {"quit", "", ""}, NULL},
  // vi_quithelp verbɽʤ
  //  {vi_score, {"score", "ȤƤ", "ȥƥ"}, NULL},
  {vi_start, {"start", "Ϥ", "ϥ"}, NULL},
  {vi_title, {"title", "Ȥ", "ȥ"}, NULL},
#ifdef DEBUG
  {vi_debug, {"debug", "ǤФä", "ǥХå"}, NULL},
  {vi_test, {"test", "Ƥ", "ƥ"}, NULL},
  {vi_teleport, {"teleport", "Ƥݡ", "ƥݡ"}, NULL},
  {vi_summon, {"summon", "礦", "祦"}, NULL},
  {vi_flag, {"flag", "դ餰", "ե饰"}, NULL},
#endif
};
#define VERB_NUM (sizeof(verbs)/sizeof(verb_t))

static esynonym_t esynonyms[] = {
    {"escape", vi_exit},
    {"search", vi_examine},
    {"destroy", vi_destroy},
    {"take", vi_get},
    {"set", vi_put},
    {"call", vi_talk},
    {"assemble", vi_combine},
    {"disassemble", vi_dismantle},
    {"climb", vi_up},
    {"bye", vi_quit},

    {"play", vi_dont_care},
};

static jsynonym_t jsynonyms[] = {
    {"ޤ", "ޥ", vi_forward},
    {"", "", vi_backward},
    {"", "", vi_up},
    {"", "", vi_down},
    {"Τܤ", "Υܥ", vi_up},
    {"", "", vi_down},
    {"Ť", "ť", vi_forward},

	{"ˤ", "˥", vi_exit},
	{"", "", vi_inventory},
	{"", "", vi_inventory},
	{"Ƥ", "ƥ", vi_inventory},
    {"Ƥˤ", "ƥ˥", vi_get},
    {"ˤ夦", "˥奦", vi_get},
    {"", "", vi_open},
    {"", "", vi_close},
    {"ҤäѤ", "ҥåѥ", vi_pull},
	{"դ", "ե", vi_touch},
	{"ʤ", "ʥ", vi_hit},
	{"ȤФ", "ȥХ", vi_kick},
	{"", "", vi_talk},
	{"", "", vi_talk},
    {"ʤǤ", "ʤǤ", vi_touch},

	{"ߤ碌", "ߥ糧", vi_combine},
	{"ߤƤ", "ߥƥ", vi_combine},
    {"Ф餹", "Х饹", vi_dismantle},
    {"Ϥ", "ϥ", vi_dismantle},
	{"ɤ", "ɥ", vi_remove},
	{"錄", "勵", vi_give},
	{"", "", vi_give},

    //	{"", "", vi_score},
	{"", "", vi_start},
	{"", "", vi_start},
	{"", "", vi_quit},
	{"夦礦", "奦祦", vi_quit},
	{"褦ʤ", "襦ʥ", vi_quit},
	{"ʤ", "ʥ", vi_quit},
	{"", "", vi_quit},
	{"ɤ", "ɥ", vi_dont_care},
	{"Ĥ", "ĥ", vi_dont_care},
	{"᤯", "᥯", vi_dont_care},
	{"󤽤", "󥽥", vi_dont_care},
#ifdef DEBUG
    {"Ƥ", "ƥ", vi_teleport},
#endif
    //    {vi_},
};

#define ESYNONYM_NUM (sizeof(esynonyms)/sizeof(esynonym_t))
#define JSYNONYM_NUM (sizeof(jsynonyms)/sizeof(jsynonym_t))
    
static verb_trans_t etrans[] = {
    {"play", ni_piano, vi_use},
    {"get", ni_telephone, vi_use},
    {"set", ni_clock, vi_adjust},
    {"shoot", ni_slingshot, vi_use},
};
static verb_trans_t jtrans[] = {
    {"Ȥ", ni_telephone, vi_use},
    {"Ǥ", ni_telephone, vi_use},
    {"Ĥ", ni_range, vi_use},
    {"Ĥ", ni_oven, vi_use},
    {"Ĥ", ni_tv, vi_use},
    {"Ĥ", ni_lantern, vi_use},
    //    {"ɤ", ni_ladder, vi_put},
    {"Ҥ", ni_piano, vi_use},
    {"󤽤", ni_piano, vi_use},
    {"Ϥ", ni_portrait, vi_remove},
    {"Ϥ", ni_screw, vi_remove},
    {"᤯", ni_mattress, vi_remove},
    {"Ҥ餯", ni_origami_plane, vi_dismantle},
    {"", ni_slingshot, vi_use},
    {"Ҥ餯", ni_wastepaper, vi_examine},
    //    {"", ni_, vi },
};
#define ETRANS_NUM (sizeof(etrans)/sizeof(verb_trans_t))
#define JTRANS_NUM (sizeof(jtrans)/sizeof(verb_trans_t))

static dic_t verb_dic;
static verb_id_t trans_table[WORD_MAX_NUM][ni_end];

int vid2verb_ix[vi_end];

static history_t verb_hist;


void verb_init(void){
    int i, j, ix;
    verb_id_t vid;

    init_dic(&verb_dic);
    init_history(&verb_hist);    
    
    // ư򼭽Ͽ
	// ƬΥߡФΤ1
    for (i = 1 ; i < VERB_NUM ; i++){

		vid = verbs[i].id;
        if (vid2verb_ix[vid] != 0){
            printf("ƥ२顼: "
                   "ư%sפID(%d)"
                   "ư%sפIDȽʣƤޤ\n", 
                   verbs[i].word[0], vid, verbs[vid2verb_ix[vid]].word[0]);
            exit(1);
        }
        //        printf("set vid:%d ix:%d\n", vid, i);
        vid2verb_ix[vid] = i;
		for (j = 0 ; j < 3 ; j++){
            register_dic(&verb_dic, verbs[i].word[j], verbs[i].id);
        }
        if (verbs[i].shortcut != NULL){
            register_dic(&verb_dic, verbs[i].shortcut, verbs[i].id);
        }
    }
    // 򼭽Ͽ
    for (i = 0 ; i < ESYNONYM_NUM ; i++){
        register_dic(&verb_dic, esynonyms[i].word, esynonyms[i].id);
    }
    for (i = 0 ; i < JSYNONYM_NUM ; i++){
        register_dic(&verb_dic, jsynonyms[i].hiragana, jsynonyms[i].id);
        register_dic(&verb_dic, jsynonyms[i].katakana, jsynonyms[i].id);
    }
    // Ṵ̣̄Ѵơ֥κ
    for (i = 0 ; i < ETRANS_NUM ; i++){
        ix = get_verb_ix(etrans[i].word);
        if (ix <= 0){
            printf("Error: trans for Unknown verb '%s'.\n", etrans[i].word);
            exit(1);
        }
        trans_table[ix][etrans[i].nid] = etrans[i].new_vid;
    }
    for (i = 0 ; i < JTRANS_NUM ; i++){
        // Ҥ餬ʤϿ
        ix = get_verb_ix(jtrans[i].word);
        if (ix < 0){
            printf("Error: trans for Unknown verb '%s'.\n", jtrans[i].word);
            exit(1);
        }
        trans_table[ix][jtrans[i].nid] = jtrans[i].new_vid;
        // ʤϿ
        vid = word_ix2id(ix, &verb_dic); // Ҥ餬ʤvid
        for (j = 1 ; j < VERB_NUM ; j++){
            if (verbs[j].id == vid){
                ix = get_verb_ix(verbs[j].word[im_katakana]);
                if (ix < 0){
                    printf("Error: trans for Unknown verb '%s'.\n",
                           verbs[j].word[im_katakana]);
                    exit(1);
                }
                break;
            }
        }
    }
    // åñ줬ҥȥ˻ĤΤǥҥȥ
    init_history(&verb_hist);
}
int get_verb_ix(char *word){
    return get_word_ix(word, &verb_dic);
}

verb_id_t get_verb_id(char *verb, noun_id_t nid){
    int ix;
    verb_id_t vid = vi_unknown, new_vid;

    ix = get_verb_ix(verb);
    register_verb_history(ix);
    if (ix >= 0){
        new_vid = trans_table[ix][nid];
        if (new_vid){
            vid = new_vid;
        }
        else {
            vid = word_ix2id(ix, &verb_dic);
        }
    }
    return vid;
}
/*
char *vid2verb(verb_id_t vid){
    int i;
    input_mode_t mode = get_config_input();

    for (i = 0 ; i < VERB_NUM ; i++){
        if (verbs[i].id == vid){
            return verbs[i].word[mode];
        }
    }
    return NULL;
}
*/
char *vid2verb(verb_id_t vid){
    input_mode_t mode = get_term_input_mode();
	int ix = vid2verb_ix[vid];

	if (ix > 0){
        return verbs[ix].word[mode];
    }
    return NULL;
}



void register_verb_history(int verb_ix){
    register_history(verb_ix, &verb_hist);
}
char *get_history_verb(void){
	history_entry_t entry = verb_hist.entries[verb_hist.ix];
	if (entry.ix == INVALID_WORD_IX) return NULL;
	return ix2word(entry.ix, &verb_dic);
};
void prev_history_verb(void){
	prev_history(&verb_hist);
}
void next_history_verb(void){
	next_history(&verb_hist);
}

 
void show_help_verb(void){
    char buf[CONSOLE_CHR_WIDTH * CONSOLE_CHR_HEIGHT], *p;
    int i;
    input_mode_t mode = get_term_input_mode();
    char *word;

    message("ʲư줬ѤǤޤ");
    buf[0] = '\0';
    p = buf;
	// ƬΥߡФΤ1
    for (i = 1 ; i < VERB_NUM ; i++){
        word = verbs[i].word[mode];
        if (verbs[i].shortcut != NULL && mode == im_english){
            //            printf("test|%s|%s|\n", word, verbs[i].shortcut);
            *p++ = verbs[i].shortcut[0];
            *p++ = '(';
            strcpy(p, word+1);
            p += strlen(word+1);
            *p++ = ')';
        }
        else {
            strcpy(p, word);
            p += strlen(word);
        }
        *p++ = ' ';
        if (verbs[i].id == vi_quit) break;


        /*
        if (verbs[i].shortcut != 0 && mode == im_english){
            *p++ = verbs[i].shortcut[0];
            *p++ = '(';
            strcpy(p, &verbs[i].word[im_english][1]);
            p += strlen(&verbs[i].word[im_english][1]);
            *p++ = ')';
        }
        */
    }
    *p = '\0';
    puts(buf);
    message(buf);
}

void show_help_basic(void){
    message("ưפ⤷ϡư ̾פηϤƤ");
    message("%v(f)ס%v(b)ס%v(l)ס%v(r)פ庸"
            "%vס%vפǾ岼˰ưǤޤ%vפǻʪǧǤޤ",
            vi_forward,
            vi_backward,
            vi_left,
            vi_right,
            vi_up, vi_down,
            vi_inventory);
    message("%vס%vס%vפѹɽ¸Ǥޤ",
    vi_config, vi_help, vi_save);
    //    message("ϡ%vס¾Υޥɤϡ%vפɽǤޤ", 
    //            vi_config, vi_help);
    //    message("вϡ%vפ¸Ǥޤ", 
    //            vi_save);
}

void show_help_tips(void){
    message("%vפǤǤϡǥȥå(ϥޤ)¸ߤޤ",
            vi_save);
	message("Ū(̾)ʤΡ%vפǡϤ򸫤뤳ȤǤޤ",
            vi_look);
    message("%vפϻ֤вᤷޤ󡣡%vפϾܤĴ٤ޤ", 
            vi_look, vi_examine);
//    message("鷺fסbסlסrפ庸˰ưǤޤ");
    message("󼨤ư졦̾ΤߤǥꥢǽǤ"
            "դޤ");
    message("ʸ񤫤줿Τϡ%v ̾פɤޤ", vi_read);
}

void show_help_key(void){
    message("ѻڡ: ޥ");
    message("꥿            : ޥɷ");
    message("[PageUp][PageDown]  : åΥХå");
    message("                : ҥȥ(Ϥư졦̾ƤӽФ)");
    message("F1                  : ");
}
void show_help_assistance(void){
    message("F2                  : ߥ˥ޥåפɽON/OFF");
    message("F3                  : ߥ˥ޥåפɽְư");
    message("F4                  : ߥ˥ޥåפƩ/ȾƩ");
    message("⤤          : ɤܤʤɤƤɽ");
}
